
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  LRU.C
*   DESCRIP  :  FENRIS LRU Module
*   DATE     :  November 19, 1998
*
*   This is how NetWare does asynch I/O with an LRU.  Somewhat different
*   than what's in Linux today.  This is a much more compact version
*   than what is in NetWare, and a little slower since it's not
*   written in hand-optimized IA32 assembly language.
*
***************************************************************************/

#include "globals.h"

extern ULONG insert_io(ULONG disk, ASYNCH_IO *io);
extern void RunAsynchIOQueue(ULONG disk);

// mirror round robin counter
ULONG mirror_counter = 0;

void NWLockLRU(LRU_HANDLE *lru_handle)
{
#if (LINUX_SLEEP)
    if (WaitOnSemaphore(&lru_handle->LRUSemaphore) == -EINTR)
       NWFSPrint("lock lru was interupted\n");
#endif
}

void NWUnlockLRU(LRU_HANDLE *lru_handle)
{
#if (LINUX_SLEEP)
    SignalSemaphore(&lru_handle->LRUSemaphore);
#endif
}

ULONG NWLockBuffer(LRU *lru)
{
#if (LINUX_SLEEP)
#if (VFS_DEBUG)
    if (down_interruptible(&lru->Semaphore) == -EINTR)
    {
       NWFSPrint("lock buffer was interupted\n");
       return -1;
    }
#else
    if (WaitOnSemaphore(&lru->Semaphore) == -EINTR)
    {
       NWFSPrint("lock buffer was interupted\n");
       return -1;
    }
#endif
#endif
    return 0;
}

void NWUnlockBuffer(LRU *lru)
{
#if (LINUX_SLEEP)
    SignalSemaphore(&lru->Semaphore);
#endif
}

void ReleaseWaiters(LRU *lru)
{
    while (lru->Waiters)
    {
#if (LINUX_SLEEP)
       SignalSemaphore(&lru->Semaphore);
#endif
       lru->Waiters--;
    }
}

#if (LINUX_20 | LINUX_22 | LINUX_24)

ULONG GetLRUTime(void)
{
    ULONG year, month, day, hour, minute, second;

    GetUnixTime(NWFSGetSystemTime(),
                &second, &minute, &hour, &day, &month, &year);
    return (ULONG)(second + (minute * 60));
}

#endif

LRU *RemoveListHead(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru_handle->ListHead == lru)
    {
       lru_handle->ListHead = (void *) lru->listnext;
       if (lru_handle->ListHead)
	  lru_handle->ListHead->listprior = NULL;
       else
	  lru_handle->ListTail = NULL;
    }
    else
    {
       lru->listprior->listnext = lru->listnext;
       if (lru != lru_handle->ListTail)
	  lru->listnext->listprior = lru->listprior;
       else
	  lru_handle->ListTail = lru->listprior;
    }
    lru->listnext = lru->listprior = 0;
    return lru;

}

void InsertListHead(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (!lru_handle->ListHead)
    {
       lru_handle->ListHead = lru;
       lru_handle->ListTail = lru;
       lru->listnext = lru->listprior = 0;
    }
    else
    {
       lru_handle->ListTail->listnext = lru;
       lru->listnext = 0;
       lru->listprior = lru_handle->ListTail;
       lru_handle->ListTail = lru;
    }
    return;

}

// this is a very SLOW brute force ordering routine.  We only use
// it when we are doing synchronous I/O (debug mode).

LRU *IndexDirtyLRU(LRU_HANDLE *lru_handle, LRU *i)
{
    register int j;
    ULONG count = 0, ccode;
    LRU *old, *p;
    nwvp_asynch_map map[8];

    for (j=0; j < 8; j++)
    {
       i->lba[j] = 0;
       i->disk[j] = 0;
    }

    // This function returns a mirror map of all mirror
    // group members, whether IN_SYNC or not.  For the
    // write case, we always attempt ot update all the
    // mirrored members asynchronously.  For the read case,
    // we only allow reads from IN_SYNC mirror members.

    ccode = nwvp_vpartition_map_asynch_write(i->nwvp_handle,
                                             i->block,
                                             &count,
					     &map[0]);
    if (!ccode && count)
    {
       i->mirror_count = (count % 8);
       for (j=0; j < (int)(count % 8); j++)
       {
          i->lba[j] = map[j].sector_offset;
          i->disk[j] = map[j].disk_id;
       }
    }

    if (!lru_handle->LRUDirtyTail)
    {
       i->dnext = i->dprior = NULL;
       lru_handle->LRUDirtyTail = i;
       return i;
    }
    p = lru_handle->LRUDirtyHead;
    old = NULL;
    while (p)
    {
       if ((p->disk < i->disk) && (p->lba[0] < i->lba[0]))
       {
	  old = p;
	  p = p->dnext;
       }
       else
       {
	  if (p->dprior)
	  {
	     p->dprior->dnext = i;
	     i->dnext = p;
	     i->dprior = p->dprior;
	     p->dprior = i;
	     return lru_handle->LRUDirtyHead;
	  }
	  i->dnext = p;
	  i->dprior = NULL;
	  p->dprior = i;
	  return i;
       }
    }
    old->dnext = i;
    i->dnext = NULL;
    i->dprior = old;
    lru_handle->LRUDirtyTail = i;
    return lru_handle->LRUDirtyHead;
}


LRU *RemoveDirty(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru_handle->LRUDirtyHead == lru)
    {
       lru_handle->LRUDirtyHead = (void *) lru->dnext;
       if (lru_handle->LRUDirtyHead)
	  lru_handle->LRUDirtyHead->dprior = NULL;
       else
	  lru_handle->LRUDirtyTail = NULL;
    }
    else
    {
       lru->dprior->dnext = lru->dnext;
       if (lru != lru_handle->LRUDirtyTail)
	  lru->dnext->dprior = lru->dprior;
       else
	  lru_handle->LRUDirtyTail = lru->dprior;
    }

    lru->dnext = lru->dprior = 0;
    lru->state &= ~L_DIRTY;
    lru->owner &= ~QUEUE_DIRTY;
    if (lru_handle->LRUDirtyBuffers)
       lru_handle->LRUDirtyBuffers--;
    return lru;

}

#if (DO_ASYNCH_IO & !DOS_UTIL & !WINDOWS_NT_UTIL & !LINUX_UTIL)

// the asynch io manager orders requests. so we don't need to do
// this here.

void InsertDirty(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru->owner & QUEUE_DIRTY)
    {
       NWFSPrint("dirty element was already on list\n");
       return;
    }

    if (!lru_handle->LRUDirtyHead)
    {
       lru_handle->LRUDirtyHead = lru;
       lru_handle->LRUDirtyTail = lru;
       lru->dnext = lru->dprior = 0;
    }
    else
    {
       lru_handle->LRUDirtyTail->dnext = lru;
       lru->dnext = 0;
       lru->dprior = lru_handle->LRUDirtyTail;
       lru_handle->LRUDirtyTail = lru;
    }
    lru->owner |= QUEUE_DIRTY;
    lru->state |= L_DIRTY;
    lru_handle->LRUDirtyBuffers++;
    return;

}

#else

void InsertDirty(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru->owner & QUEUE_DIRTY)
    {
       NWFSPrint("dirty element was already on list\n");
       return;
    }
    lru_handle->LRUDirtyHead = IndexDirtyLRU(lru_handle, lru);
    lru->owner |= QUEUE_DIRTY;
    lru->state |= L_DIRTY;
    lru_handle->LRUDirtyBuffers++;
    return;
}

#endif

LRU *RemoveLRU(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru_handle->LRUListHead == lru)
    {
       lru_handle->LRUListHead = (void *) lru->next;
       if (lru_handle->LRUListHead)
	  lru_handle->LRUListHead->prior = NULL;
       else
	  lru_handle->LRUListTail = NULL;
    }
    else
    {
       lru->prior->next = lru->next;
       if (lru != lru_handle->LRUListTail)
	  lru->next->prior = lru->prior;
       else
	  lru_handle->LRUListTail = lru->prior;
    }

    lru->next = lru->prior = 0;
    lru->owner &= ~QUEUE_LRU;
    return lru;

}

void InsertLRU(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru->owner & QUEUE_LRU)
    {
       NWFSPrint("LRU element was already on list\n");
       return;
    }

    if (!lru_handle->LRUListHead)
    {
       lru_handle->LRUListHead = lru;
       lru_handle->LRUListTail = lru;
       lru->next = lru->prior = 0;
    }
    else
    {
       lru_handle->LRUListTail->next = lru;
       lru->next = 0;
       lru->prior = lru_handle->LRUListTail;
       lru_handle->LRUListTail = lru;
    }
    lru->owner |= QUEUE_LRU;
    return;

}

void InsertLRUTop(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru->owner & QUEUE_LRU)
    {
       NWFSPrint("LRU element was already on list\n");
       return;
    }

    if (!lru_handle->LRUListHead)
    {
       lru_handle->LRUListHead = lru;
       lru_handle->LRUListTail = lru;
       lru->next = lru->prior = 0;
    }
    else
    {
       lru->next = lru_handle->LRUListHead;
       lru->prior = 0;
       lru_handle->LRUListHead->prior = lru;
       lru_handle->LRUListHead = lru;
    }
    lru->owner |= QUEUE_LRU;
    return;

}

LRU *GetFreeLRU(LRU_HANDLE *lru_handle)
{
    register LRU *lru;

    if (lru_handle->LRUFreeHead)
    {
       lru = lru_handle->LRUFreeHead;
       lru_handle->LRUFreeHead = (void *) lru->next;
       if (lru_handle->LRUFreeHead)
	  lru_handle->LRUFreeHead->prior = NULL;
       else
	  lru_handle->LRUFreeTail = NULL;

       lru->next = lru->prior = 0;
       lru->owner &= ~QUEUE_FREE;
       return lru;
    }
    return 0;

}

void PutFreeLRU(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru->owner & QUEUE_FREE)
    {
       NWFSPrint("free element was already on list\n");
       return;
    }

    if (!lru_handle->LRUFreeHead)
    {
       lru_handle->LRUFreeHead = lru;
       lru_handle->LRUFreeTail = lru;
       lru->next = lru->prior = 0;
    }
    else
    {
       lru_handle->LRUFreeTail->next = lru;
       lru->next = 0;
       lru->prior = lru_handle->LRUFreeTail;
       lru_handle->LRUFreeTail = lru;
    }
    lru->owner |= QUEUE_FREE;
    return;

}

LRU *SearchHash(LRU_HANDLE *lru_handle, VOLUME *volume, ULONG block)
{
    register LRU *lru;
    register ULONG hash;
    register DATA_LRU_HASH_LIST *HashTable;

    hash = (block & (lru_handle->LRUBlockHashLimit - 1));
    HashTable = (DATA_LRU_HASH_LIST *) lru_handle->LRUBlockHash;
    if (HashTable)
    {
       lru = HashTable[hash].head;
       while (lru)
       {
	  if ((lru->block == block) && (lru->volnum == volume->VolumeNumber))
	     return lru;
	  lru = lru->hashNext;
       }
    }
    return 0;
}

ULONG AddToHash(LRU_HANDLE *lru_handle, LRU *lru)
{
    register ULONG hash;
    register DATA_LRU_HASH_LIST *HashTable;

    hash = (lru->block & (lru_handle->LRUBlockHashLimit - 1));
    HashTable = (DATA_LRU_HASH_LIST *) lru_handle->LRUBlockHash;
    if (HashTable)
    {
       if (lru->owner & QUEUE_HASH)
       {
	  NWFSPrint("hash element was already on list\n");
	  return -1;
       }

       if (!HashTable[hash].head)
       {
	  HashTable[hash].head = lru;
	  HashTable[hash].tail = lru;
	  lru->hashNext = lru->hashPrior = 0;
       }
       else
       {
	  HashTable[hash].tail->hashNext = lru;
	  lru->hashNext = 0;
	  lru->hashPrior = HashTable[hash].tail;
	  HashTable[hash].tail = lru;
       }
       lru->owner |= QUEUE_HASH;
       return 0;
    }
    return -1;
}

ULONG RemoveHash(LRU_HANDLE *lru_handle, LRU *lru)
{
    register ULONG hash;
    register DATA_LRU_HASH_LIST *HashTable;

    hash = (lru->block & (lru_handle->LRUBlockHashLimit - 1));
    HashTable = (DATA_LRU_HASH_LIST *) lru_handle->LRUBlockHash;
    if (HashTable)
    {
       if (HashTable[hash].head == lru)
       {
	  HashTable[hash].head = lru->hashNext;
	  if (HashTable[hash].head)
	     HashTable[hash].head->hashPrior = NULL;
	  else
	     HashTable[hash].tail = NULL;
       }
       else
       {
	  lru->hashPrior->hashNext = lru->hashNext;
	  if (lru != HashTable[hash].tail)
	     lru->hashNext->hashPrior = lru->hashPrior;
	  else
	     HashTable[hash].tail = lru->hashPrior;
       }

       lru->hashNext = lru->hashPrior = 0;
       lru->owner &= ~QUEUE_HASH;
       return 0;
    }
    return -1;
}

void FreeLRUElement(LRU_HANDLE *lru_handle, LRU *lru)
{
    if (lru)
    {
       if (lru->buffer)
       {
#if (LINUX_20 | LINUX_22)
     	  kfree(lru->buffer);
#elif (LINUX_24)
          __free_page(lru->page);
#else
	  NWFSFree(lru->buffer);
#endif
       }
       NWFSFree(lru);
       if (lru_handle->LRUBlocks)
	  lru_handle->LRUBlocks--;
    }
    return;
}

LRU *AllocateLRUElement(LRU_HANDLE *lru_handle)
{
    register LRU *lru;

    lru = NWFSAlloc(sizeof(LRU), LRU_TAG);
    if (!lru)
       return 0;

    NWFSSet(lru, 0, sizeof(LRU));
    lru->state = L_FREE;
    lru->block = -1;
    lru->volnum = (ULONG) -1;
    lru->size = IO_BLOCK_SIZE;
    lru->lru_handle = lru_handle;

#if (LINUX_SLEEP)
    AllocateSemaphore(&lru->Semaphore, 0);  // this is a waiters queue
#endif

    // if we are a linux driver, then the buffer must be page aligned
    // or the lower layers may get DMA timeouts (drivers in Linux are
    // optimized for 512-byte aligned memory).

#if (LINUX_20 | LINUX_22)
    lru->buffer = kmalloc(IO_BLOCK_SIZE, GFP_BUFFER);
    if (!lru->buffer)
    {
       NWFSFree(lru);
       return 0;
    }
#elif (LINUX_24)
    lru->page = alloc_page(GFP_BUFFER);
    if (!lru->page)
    {
       NWFSFree(lru);
       return 0;
    }
    lru->buffer = (BYTE *)page_address(lru->page);
#else
    lru->buffer = NWFSIOAlloc(IO_BLOCK_SIZE, LRU_BUFFER_TAG);
    if (!lru->buffer)
    {
       NWFSFree(lru);
       return 0;
    }
#endif

    NWFSSet(lru->buffer, 0, IO_BLOCK_SIZE);
    lru_handle->LRUBlocks++;

    InsertListHead(lru_handle, lru);
    return lru;
}

ULONG InitializeLRU(LRU_HANDLE *lru_handle, BYTE *name, ULONG mode,
		    ULONG max, ULONG min)
{
    NWFSSet(lru_handle, 0, sizeof(LRU_HANDLE));
    lru_handle->LRUName = name;
    lru_handle->LRUMode = mode;

#if (LINUX_SLEEP)
    AllocateSemaphore(&lru_handle->LRUSemaphore, 1);
    AllocateSemaphore(&lru_handle->LRUSyncSemaphore, 1);
#endif

    lru_handle->LRUBlockHash = NWFSCacheAlloc(BLOCK_NUMBER_HASH_SIZE, LRU_HASH_TAG);
    if (!lru_handle->LRUBlockHash)
       return -1;

    lru_handle->LRUBlockHashLimit = NUMBER_OF_BLOCK_HASH_ENTRIES;
    NWFSSet(lru_handle->LRUBlockHash, 0, BLOCK_NUMBER_HASH_SIZE);

    lru_handle->MinimumLRUBlocks = min;
    lru_handle->MaximumLRUBlocks = max;
    lru_handle->PERCENT_FACTOR = LRU_FACTOR;
    lru_handle->AGING_INTERVAL = LRU_AGE;

#if (LINUX_SLEEP)
    lru_handle->LRUWaiters = 0;
    AllocateSemaphore(&lru_handle->LRUWaiterSemaphore, 0);  // this is a waiters queue
#endif

    return 0;
}

ULONG FreeLRU(LRU_HANDLE *lru_handle)
{
    register ULONG ccode;
    register LRU *lru, *list;

    NWLockLRU(lru_handle);
    lru_handle->Active = TRUE;
    list = lru_handle->ListHead;
    lru_handle->ListHead = lru_handle->ListTail = 0;
    while (list)
    {
       lru = list;
       list = list->listnext;

       if ((lru->state & L_DIRTY)      &&
	   (lru->state & L_DATAVALID)  &&
	   (lru->nwvp_handle)          &&
	   (!(lru->state & L_FLUSHING)))
       {
	  lru->state |= L_FLUSHING;
	  lru->state |= L_UPTODATE;
          NWUnlockLRU(lru_handle);

	  ccode = nwvp_vpartition_block_write(lru->nwvp_handle, lru->block, 1, lru->buffer);

          NWLockLRU(lru_handle);
	  lru->state &= ~L_FLUSHING;
	  lru->timestamp = 0;

#if (VERBOSE)
	  if (ccode)
	     NWFSPrint("nwfs:  I/O Error (close) block-%d volume-%d\n",
		       (int)lru->block, (int)lru->volnum);
#endif

	  if (lru->state & L_UPTODATE)
	     RemoveDirty(lru_handle, lru);
       }
       FreeLRUElement(lru_handle, lru);
    }

    list = lru_handle->LRUFreeHead;
    lru_handle->LRUFreeHead = lru_handle->LRUFreeTail = 0;
    while (list)
    {
       lru = list;
       list = list->next;
       FreeLRUElement(lru_handle, lru);
    }
    lru_handle->Active = FALSE;
    NWUnlockLRU(lru_handle);

    if (lru_handle->LRUBlockHash)
       NWFSFree(lru_handle->LRUBlockHash);
    NWFSSet(lru_handle, 0, sizeof(LRU_HANDLE));

    return 0;
}

ULONG release_write_aio(ASYNCH_IO *io)
{
    register ULONG ccode;
    register LRU *lru = (LRU *) io->call_back_parameter;
    register LRU_HANDLE *lru_handle = (LRU_HANDLE *) lru->lru_handle;

    NWLockLRU(lru_handle);
    if (io->ccode)
       lru->aio_ccode = io->ccode;

    // count the number of aoi's completed.  signal when the
    // last mirror member has returned completed.

    lru->mirrors_completed++;
    if (lru->mirror_count == lru->mirrors_completed)
    {
       lru->state &= ~L_FLUSHING;
       lru->timestamp = 0;

       if (!lru->aio_ccode)
       {
	  // see if someone dirtied this buffer again while asynch I/O was
	  // pending.   If they did, don't remove the buffer from the dirty
          // list.
	  if (lru->state & L_UPTODATE)
	     RemoveDirty(lru_handle, lru);
       }
       else
       {
	  // if we had a write failure for any of the mirrored asynch
	  // writes, then we retry the write operation synchronously with
	  // retry logic to attempt to recover lost sectors (hotfixing).
	  // nwvp_() function calls perform hotfixing and mirror failover
	  // operations.

	  lru->state |= L_UPTODATE;
	  NWUnlockLRU(lru_handle);

	  ccode = nwvp_vpartition_block_write(lru->nwvp_handle, lru->block, 1, lru->buffer);

	  NWLockLRU(lru_handle);
	  lru->state &= ~L_FLUSHING;
	  lru->timestamp = 0;

	  // if no I/O error and the buffer was not dirtied while I/O was
          // pending, then remove from the dirty list.
	  if (!ccode && (lru->state & L_UPTODATE))
	     RemoveDirty(lru_handle, lru);
       }
    }
    NWUnlockLRU(lru_handle);
    return 0;
}

ULONG FlushLRU(LRU_HANDLE *lru_handle)
{
#if (DO_ASYNCH_IO & !DOS_UTIL & !WINDOWS_NT_UTIL & !LINUX_UTIL)

    register ULONG i;
    register LRU *lru, *list;
    BYTE disks[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
    register int j;
    ULONG count = 0, ccode;
    nwvp_asynch_map map[8];

    NWLockLRU(lru_handle);

    lru_handle->Active = TRUE;
    list = lru_handle->LRUDirtyHead;
    while (list)
    {
       lru = list;
       list = list->dnext;

       if ((lru->state & L_DIRTY)      &&
	   (lru->state & L_DATAVALID)  &&
	   (lru->nwvp_handle)          &&
           (!lru->lock_count)          &&
	   (!(lru->state & L_FLUSHING)))
       {
	  lru->state |= L_FLUSHING;

	  // this flag is cleared when we dirty a buffer
	  lru->state |= L_UPTODATE;
	  NWUnlockLRU(lru_handle);

	  NWLockLRU(lru_handle);
          for (j=0; j < 8; j++)
          {
             lru->lba[j] = 0;
             lru->disk[j] = 0;
          }

          // This function returns a mirror map of all mirror
          // group members, whether IN_SYNC or not.  For the
          // write case, we always attempt ot update all the
          // mirrored members asynchronously.  For the read case,
          // we only allow reads from IN_SYNC mirror members.

          ccode = nwvp_vpartition_map_asynch_write(lru->nwvp_handle,
                                                   lru->block,
                                                   &count,
					           &map[0]);
          if (!ccode && count)
          {
             lru->mirror_count = (count % 8);
             for (j=0; j < (int)(count % 8); j++)
             {
                lru->lba[j] = map[j].sector_offset;
                lru->disk[j] = map[j].disk_id;
             }
          }

	  lru->mirrors_completed = 0;
	  for (i=0; i < lru->mirror_count; i++)
          {
             NWFSSet(&lru->io[i], 0, sizeof(ASYNCH_IO));
	     lru->io[i].command = ASYNCH_WRITE;
	     lru->io[i].disk = lru->disk[i];
             lru->io[i].sector_offset = lru->lba[i];
             lru->io[i].sector_count = 8;
             lru->io[i].buffer = lru->buffer;
	     lru->io[i].call_back_routine = release_write_aio;
             lru->io[i].call_back_parameter = (ULONG)lru;
	     
             // the file system is not allowed to ever write sector
             // zero at this layer.
             if (lru->io[i].sector_offset == 0)
             {
                NWFSPrint("nwfs:  mirroring map is corrupted block-%d vol-%d.  Aborting I/O\n",
		         (int)lru->block, (int)lru->volnum);
                continue;
             }
             insert_io(lru->disk[i], &lru->io[i]);
             disks[(lru->disk[i] % 8)] = TRUE;
          }
       }
    }

    // signal disk processes for any queues we gave work to.
    for (i=0; i < 8; i++)
    {
       if (disks[i])
	  RunAsynchIOQueue(i);
    }

    lru_handle->Active = FALSE;
    NWUnlockLRU(lru_handle);
    return 0;

#else

    register ULONG ccode;
    register LRU *lru, *list;

    NWLockLRU(lru_handle);
    lru_handle->Active = TRUE;
    list = lru_handle->LRUDirtyHead;
    while (list)
    {
       lru = list;
       list = list->dnext;

       if ((lru->state & L_DIRTY)      &&
	   (lru->state & L_DATAVALID)  &&
	   (lru->nwvp_handle)          &&
	   (!lru->lock_count)          &&
	   (!(lru->state & L_FLUSHING)))
       {
	  lru->state |= L_FLUSHING;

	  // this flag is cleared when we dirty a buffer
	  lru->state |= L_UPTODATE;
          NWUnlockLRU(lru_handle);

	  ccode = nwvp_vpartition_block_write(lru->nwvp_handle, lru->block, 1, lru->buffer);

          NWLockLRU(lru_handle);
	  lru->state &= ~L_FLUSHING;
	  lru->timestamp = 0;

	  if (!ccode && (lru->state & L_UPTODATE))
	     RemoveDirty(lru_handle, lru);
       }
    }
    lru_handle->Active = FALSE;
    NWUnlockLRU(lru_handle);
    return 0;

#endif
}

ULONG FlushEligibleLRU(LRU_HANDLE *lru_handle)
{
#if (DO_ASYNCH_IO & !DOS_UTIL & !WINDOWS_NT_UTIL & !LINUX_UTIL)

    register ULONG i;
    register LRU *lru, *list;
    BYTE disks[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
    register int j;
    ULONG count = 0, ccode;
    nwvp_asynch_map map[8];

    NWLockLRU(lru_handle);
    lru_handle->Active = TRUE;
    list = lru_handle->LRUDirtyHead;
    while (list)
    {
       lru = list;
       list = list->dnext;

#if (LINUX_20 | LINUX_22 | LINUX_24)
       if ((lru_handle->State != NWFS_FLUSH_ALL) &&
           (lru->timestamp + lru_handle->AGING_INTERVAL) > GetLRUTime())
	  continue;
#endif
       if ((lru->state & L_DIRTY)      &&
           (lru->state & L_DATAVALID)  &&
	   (lru->nwvp_handle)          &&
           (!lru->lock_count)          &&
	   (!(lru->state & L_FLUSHING)))
       {
	  lru->state |= L_FLUSHING;

	  // this flag is cleared when we dirty a buffer
	  lru->state |= L_UPTODATE;
          NWUnlockLRU(lru_handle);

          NWLockLRU(lru_handle);
          for (j=0; j < 8; j++)
          {
             lru->lba[j] = 0;
             lru->disk[j] = 0;
          }

          // This function returns a mirror map of all mirror
          // group members, whether IN_SYNC or not.  For the
          // write case, we always attempt ot update all the
          // mirrored members asynchronously.  For the read case,
          // we only allow reads from IN_SYNC mirror members.

          ccode = nwvp_vpartition_map_asynch_write(lru->nwvp_handle,
                                                   lru->block,
                                                   &count,
					           &map[0]);
          if (!ccode && count)
          {
             lru->mirror_count = (count % 8);
             for (j=0; j < (int)(count % 8); j++)
             {
                lru->lba[j] = map[j].sector_offset;
                lru->disk[j] = map[j].disk_id;
             }
          }

	  lru->mirrors_completed = 0;
	  for (i=0; i < lru->mirror_count; i++)
          {
             NWFSSet(&lru->io[i], 0, sizeof(ASYNCH_IO));
             lru->io[i].command = ASYNCH_WRITE;
             lru->io[i].disk = lru->disk[i];
             lru->io[i].sector_offset = lru->lba[i];
	     lru->io[i].sector_count = 8;
	     lru->io[i].buffer = lru->buffer;
	     lru->io[i].call_back_routine = release_write_aio;
             lru->io[i].call_back_parameter = (ULONG)lru;
	     
             // the file system is not allowed to ever write sector
             // zero at this layer.
             if (lru->io[i].sector_offset == 0)
             {
                NWFSPrint("nwfs:  mirroring map is corrupted block-%d vol-%d.  Aborting I/O\n",
		         (int)lru->block, (int)lru->volnum);
                continue;
             }
             insert_io(lru->disk[i], &lru->io[i]);
             disks[(lru->disk[i] % 8)] = TRUE;
          }
       }
    }

    // signal disk processes for any queues we gave work to.
    for (i=0; i < 8; i++)
    {
       if (disks[i])
          RunAsynchIOQueue(i);
    }
    lru_handle->State = 0;
    lru_handle->Active = FALSE;
    NWUnlockLRU(lru_handle);
    return 0;

#else

    register ULONG ccode;
    register LRU *lru, *list;

    NWLockLRU(lru_handle);
    lru_handle->Active = TRUE;
    list = lru_handle->LRUDirtyHead;
    while (list)
    {
       lru = list;
       list = list->dnext;

#if (LINUX_20 | LINUX_22 | LINUX_24)
       if ((lru_handle->State != NWFS_FLUSH_ALL) &&
           (lru->timestamp + lru_handle->AGING_INTERVAL) > GetLRUTime())
	  continue;
#endif
       if ((lru->state & L_DIRTY)      &&
           (lru->state & L_DATAVALID)  &&
	   (lru->nwvp_handle)          &&
           (!lru->lock_count)          &&
	   (!(lru->state & L_FLUSHING)))
       {
	  lru->state |= L_FLUSHING;
	  lru->state |= L_UPTODATE;
          NWUnlockLRU(lru_handle);

	  ccode = nwvp_vpartition_block_write(lru->nwvp_handle, lru->block, 1, lru->buffer);

          NWLockLRU(lru_handle);
	  lru->state &= ~L_FLUSHING;
	  lru->timestamp = 0;

	  if (!ccode && (lru->state & L_UPTODATE))
	     RemoveDirty(lru_handle, lru);
       }
    }
    lru_handle->State = 0;
    lru_handle->Active = FALSE;
    NWUnlockLRU(lru_handle);
    return 0;

#endif
}

ULONG FlushVolumeLRU(LRU_HANDLE *lru_handle, VOLUME *volume)
{
    register ULONG ccode;
    register LRU *lru, *list;

    NWLockLRU(lru_handle);
    lru_handle->Active = TRUE;
    list = lru_handle->ListHead;
    while (list)
    {
       lru = list;
       list = list->listnext;

       if (lru->volnum == volume->VolumeNumber)
       {
	  if ((lru->state & L_DIRTY)        &&
	      (lru->state & L_DATAVALID)    &&
	      (lru->nwvp_handle)            &&
	      (!(lru->state & L_FLUSHING)))
	  {
	     lru->state |= L_FLUSHING;
	     lru->state |= L_UPTODATE;
             NWUnlockLRU(lru_handle);

	     ccode = nwvp_vpartition_block_write(lru->nwvp_handle, lru->block, 1, lru->buffer);

             NWLockLRU(lru_handle);
	     lru->state &= ~L_FLUSHING;
	     lru->timestamp = 0;

             if (lru->state & L_UPTODATE)
             {
	        RemoveDirty(lru_handle, lru);
             }
	  }
	  RemoveLRU(lru_handle, lru);
	  RemoveHash(lru_handle, lru);

	  RemoveListHead(lru_handle, lru);
	  lru->state = L_FREE;
	  lru->volnum = (ULONG) -1;
	  PutFreeLRU(lru_handle, lru);
       }
    }
    lru_handle->Active = FALSE;
    NWUnlockLRU(lru_handle);
    return 0;
}

//
//   blocking read functions
//

#if (DO_ASYNCH_IO)
ULONG release_read_aio(ASYNCH_IO *io)
{
#if (LINUX_SLEEP)
    SignalSemaphore((struct semaphore *)io->call_back_parameter);
#endif
    return 0;
}
#endif

LRU *FillBlock(LRU_HANDLE *lru_handle, VOLUME *volume, ULONG block,
	       ULONG mirror_block, ULONG fill, ULONG ccount)
{
    register ULONG retCode;
    register LRU *lru = 0;
#if (PRIORITY_LRU)
    register LRU *lru_post;
#endif

    if (lru_handle->LRUFreeHead)
    {
       lru = GetFreeLRU(lru_handle);
       if (lru)
       {
          InsertListHead(lru_handle, lru);
          goto found_lru_buffer;
       }
    }

    if (lru_handle->LRUBlocks < lru_handle->MaximumLRUBlocks)
    {
       lru = AllocateLRUElement(lru_handle);
       if (lru)
          goto found_lru_buffer;

       lru_handle->MaximumLRUBlocks = lru_handle->LRUBlocks;
    }

    if (lru_handle->LRUListHead)
    {
       lru = lru_handle->LRUListTail;
       while ((lru) && 
             ((lru->state & L_DIRTY) || (lru->state & L_LOADING) ||
	      (lru->state & L_FLUSHING) || (lru->lock_count))) 
          lru = lru->prior;
	  
       if (lru)
       {
          RemoveLRU(lru_handle, lru);
          RemoveHash(lru_handle, lru);
       }
    }

    if (!lru)
       return 0;

found_lru_buffer:;
    lru->state = L_LOADING;
    lru->block = block;
    lru->mirror_block = mirror_block;
    lru->nwvp_handle = volume->nwvp_handle;
    lru->volnum = volume->VolumeNumber;
    lru->cycle_count = ccount;
    lru->current_cycle_count = 0;

    AddToHash(lru_handle, lru);
    InsertLRUTop(lru_handle, lru);

    if (fill)
    {
#if (DO_ASYNCH_IO & !DOS_UTIL & !WINDOWS_NT_UTIL & !LINUX_UTIL)
#if (LINUX_SLEEP)
       NWFSInitSemaphore(semaphore);
#endif
       register int j;
       ULONG count = 0, map_ccode;
       nwvp_asynch_map map[8];

       // remap the lru mirroring information since during recovery, it
       // may have changed.

       for (j=0; j < 8; j++)
       {
	  lru->lba[j] = 0;
	  lru->disk[j] = 0;
       }

       map_ccode = nwvp_vpartition_map_asynch_read(lru->nwvp_handle,
						   lru->block,
						   &count, &map[0]);
       if (!map_ccode && count)
       {
	  lru->mirror_count = (count % 8);
	  for (j=0; j < (int)(count % 8); j++)
	  {
	     lru->lba[j] = map[j].sector_offset;
	     lru->disk[j] = map[j].disk_id;
	  }
       }

       NWFSSet(&lru->io[0], 0, sizeof(ASYNCH_IO));
       lru->io[0].command = ASYNCH_READ;

       // Pick a random mirror member to read from that is IN_SYNC.
       // nwvp_vpartition_map_asynch_read() will only return a list
       // of IN_SYNC mirrors to read from, as opposed to
       // nwvp_vpartition_map_asynch_write() which will return all
       // present mirror group members with an active device.

       lru->io[0].disk = lru->disk[(mirror_counter % lru->mirror_count)];
       lru->io[0].sector_offset = lru->lba[(mirror_counter % lru->mirror_count)];
       lru->io[0].sector_count = 8;
       lru->io[0].buffer = lru->buffer;
       lru->io[0].call_back_routine = release_read_aio;
       lru->io[0].call_back_parameter = (ULONG)&semaphore;
	     
       insert_io(lru->io[0].disk, &lru->io[0]);
       mirror_counter++;

       RunAsynchIOQueue(lru->io[0].disk);

       NWUnlockLRU(lru_handle);

#if (LINUX_SLEEP)
       WaitOnSemaphore(&semaphore);
#endif
       // if we get an I/O error on the async read, retry the operation
       // synchronously to perform hotfixing or mirrored failover.

       retCode = lru->io[0].ccode;
       if (retCode)
	  retCode = nwvp_vpartition_block_read(lru->nwvp_handle, lru->block, 1, lru->buffer);

       NWLockLRU(lru_handle);
#else
       NWUnlockLRU(lru_handle);

       retCode = nwvp_vpartition_block_read(lru->nwvp_handle, lru->block, 1, lru->buffer);

       NWLockLRU(lru_handle);
#endif
       if (retCode)
       {
	  RemoveLRU(lru_handle, lru);
	  RemoveHash(lru_handle, lru);

	  RemoveListHead(lru_handle, lru);
	  lru->state = L_FREE;
	  lru->volnum = (ULONG) -1;
	  PutFreeLRU(lru_handle, lru);
	  ReleaseWaiters(lru);
	  return 0;
       }
    }
    lru->state &= ~L_LOADING;
    lru->state |= L_DATAVALID;
    ReleaseWaiters(lru);
    return lru;

}

//
//  Asynch Read Functions
//

ULONG release_asynch_read_ahead(ASYNCH_IO *io)
{
    register ULONG ccode;
    register LRU *lru = (LRU *) io->call_back_parameter;
    register LRU_HANDLE *lru_handle = (LRU_HANDLE *) lru->lru_handle;

    // if we get an I/O error on the async read, retry the operation
    // synchronously to perform hotfixing or mirrored failover.

    NWLockLRU(lru_handle);
    ccode = io->ccode;
    if (ccode)
    {
       NWUnlockLRU(lru_handle);
       ccode = nwvp_vpartition_block_read(lru->nwvp_handle, lru->block, 1, lru->buffer);
       NWLockLRU(lru_handle);
    }

    if (ccode)
    {
       RemoveLRU(lru_handle, lru);
       RemoveHash(lru_handle, lru);

       RemoveListHead(lru_handle, lru);
       lru->state = L_FREE;
       lru->volnum = (ULONG) -1;
       PutFreeLRU(lru_handle, lru);

       ReleaseWaiters(lru);
       NWUnlockLRU(lru_handle);
       return 0;
    }
    lru->state &= ~L_LOADING;
    lru->state |= L_DATAVALID;
    ReleaseWaiters(lru);
    NWUnlockLRU(lru_handle);
    return 0;

}

LRU *FillBlockAsynch(LRU_HANDLE *lru_handle, VOLUME *volume, ULONG block,
		     ULONG mirror_block, ULONG ccount, BYTE *disks)
{
    register LRU *lru = 0;
#if (PRIORITY_LRU)
    register LRU *lru_post;
#endif
    register int j;
    ULONG count = 0, map_ccode;
    nwvp_asynch_map map[8];

    if (lru_handle->LRUFreeHead)
    {
       lru = GetFreeLRU(lru_handle);
       if (lru)
       {
          InsertListHead(lru_handle, lru);
          goto found_lru_buffer;
       }
    }

    if (lru_handle->LRUBlocks < lru_handle->MaximumLRUBlocks)
    {
       lru = AllocateLRUElement(lru_handle);
       if (lru)
          goto found_lru_buffer;

       lru_handle->MaximumLRUBlocks = lru_handle->LRUBlocks;
    }

    if (lru_handle->LRUListHead)
    {
       lru = lru_handle->LRUListTail;
       while ((lru) && 
             ((lru->state & L_DIRTY) || (lru->state & L_LOADING) ||
	      (lru->state & L_FLUSHING) || (lru->lock_count))) 
          lru = lru->prior;
	  
       if (lru)
       {
          RemoveLRU(lru_handle, lru);
          RemoveHash(lru_handle, lru);
       }
    }

    if (!lru)
       return 0;

found_lru_buffer:;
    lru->state = L_LOADING;
    lru->block = block;
    lru->mirror_block = mirror_block;
    lru->nwvp_handle = volume->nwvp_handle;
    lru->volnum = volume->VolumeNumber;
    lru->cycle_count = ccount;
    lru->current_cycle_count = 0;

    AddToHash(lru_handle, lru);
    InsertLRUTop(lru_handle, lru);

    for (j=0; j < 8; j++)
    {
       lru->lba[j] = 0;
       lru->disk[j] = 0;
    }

    map_ccode = nwvp_vpartition_map_asynch_read(lru->nwvp_handle,
						lru->block,
						&count, &map[0]);
    if (!map_ccode && count)
    {
       lru->mirror_count = (count % 8);
       for (j=0; j < (int)(count % 8); j++)
       {
	  lru->lba[j] = map[j].sector_offset;
	  lru->disk[j] = map[j].disk_id;
       }
    }

    NWFSSet(&lru->io[0], 0, sizeof(ASYNCH_IO));
    lru->io[0].command = ASYNCH_READ;
    lru->io[0].disk = lru->disk[(mirror_counter % lru->mirror_count)];
    lru->io[0].sector_offset = lru->lba[(mirror_counter % lru->mirror_count)];
    lru->io[0].sector_count = 8;
    lru->io[0].buffer = lru->buffer;
    lru->io[0].call_back_routine = release_asynch_read_ahead;
    lru->io[0].call_back_parameter = (ULONG)lru;
	     
    insert_io(lru->io[0].disk, &lru->io[0]);
    mirror_counter++;

    disks[(lru->io[0].disk % 8)] = TRUE;
    return lru;
}

ULONG ReleaseLRU(LRU_HANDLE *lru_handle, LRU *lru)
{
    NWLockLRU(lru_handle);

    lru->lock_count--;
    RemoveLRU(lru_handle, lru);
    InsertLRUTop(lru_handle, lru);

    NWUnlockLRU(lru_handle);
    return 0;
}

ULONG ReleaseDirtyLRU(LRU_HANDLE *lru_handle, LRU *lru)
{
    NWLockLRU(lru_handle);

    lru->state &= ~L_UPTODATE;
    lru->lock_count--;

#if (LINUX_20 | LINUX_22 | LINUX_24)
    lru->timestamp = GetLRUTime();
#endif
    if (!(lru->state & L_DIRTY))
       InsertDirty(lru_handle, lru);

    RemoveLRU(lru_handle, lru);
    InsertLRUTop(lru_handle, lru);

    NWUnlockLRU(lru_handle);
    return 0;
}

#if (READ_AHEAD_ON)

ULONG PerformBlockReadAhead(LRU_HANDLE *lru_handle, VOLUME *volume,
			    ULONG block, ULONG blockReadAhead, ULONG ccount)
{
    register ULONG i;
    register LRU *lru;
#if (ASYNCH_READ_AHEAD && !DOS_UTIL && !WINDOWS_NT_UTIL && !LINUX_UTIL)
    BYTE disks[8] = { 0, 0, 0, 0, 0, 0, 0, 0 };
#endif

    if (!blockReadAhead)
       return -1;

    NWLockLRU(lru_handle);
    for (i = 0; i < blockReadAhead; i++)
    {
       lru = SearchHash(lru_handle, volume, block + i);
       if (!lru)
#if (ASYNCH_READ_AHEAD && !DOS_UTIL && !WINDOWS_NT_UTIL && !LINUX_UTIL)
	  FillBlockAsynch(lru_handle, volume, block + i, -1, ccount, disks);
#else
	  FillBlock(lru_handle, volume, block + i, -1, 1, ccount);
#endif
    }

#if (ASYNCH_READ_AHEAD && !DOS_UTIL && !WINDOWS_NT_UTIL && !LINUX_UTIL)
    for (i=0; i < 8; i++)
    {
       if (disks[i])
          RunAsynchIOQueue(i);
    }
#endif

    NWUnlockLRU(lru_handle);
    return 0;

}

#endif

LRU *ReadLRU(LRU_HANDLE *lru_handle, VOLUME *volume, ULONG block, ULONG
	     mirror_block, ULONG fill, ULONG ccount, ULONG readAhead)
{
#if (DOS_UTIL | LINUX_UTIL | WINDOWS_NT_UTIL)
    register ULONG retries = 0;
#endif
    register LRU *lru;

    // validate the priority of this request and adjust it to something
    // reasonable if we are passed an out of range value.
    // don't allow blocks to recycle more than three times prior to
    // being evicted from the LRU.

    if (!ccount || (ccount > 3))
       ccount = 1;

#if (READ_AHEAD_ON)
    // perform read ahead if we were asked to fill any blocks.
    if (fill && readAhead)
       PerformBlockReadAhead(lru_handle, volume, block + 1, readAhead - 1,
                             ccount);

#endif

repeat:;
    NWLockLRU(lru_handle);

    lru = SearchHash(lru_handle, volume, block);
    if (lru)
    {
       if (lru->state & L_LOADING)
	  goto WaitOnBuffer;

       // this indicates an I/O error
       if (!(lru->state & L_DATAVALID))
       {
	   NWUnlockLRU(lru_handle);
           return 0;
       }

       lru->lock_count++;
       NWUnlockLRU(lru_handle);
       return lru;
    }

#if (LINUX_20 | LINUX_22 | LINUX_24)
    if ((lru_handle->LRUDirtyBuffers * lru_handle->PERCENT_FACTOR) >=
         lru_handle->LRUBlocks)
    {
       if (!lru_handle->Signal)
       {
          if (!lru_handle->Active)
          {
	     lru_handle->Signal = TRUE;
	     NWFSFlush();
          }
       }
    }
#endif

    // this call is blocking.
    lru = FillBlock(lru_handle, volume, block, mirror_block, fill, ccount);
    if (lru)
    {
       lru->lock_count++;
       NWUnlockLRU(lru_handle);
       return lru;
    }

    NWUnlockLRU(lru_handle);

#if (LINUX_20 | LINUX_22 | LINUX_24)
    lru_handle->State = NWFS_FLUSH_ALL;
    if (!lru_handle->Signal)
    {
       if (!lru_handle->Active)
       {
	  lru_handle->Signal = TRUE;
          lru_handle->State = NWFS_FLUSH_ALL;
	  NWFSFlush();
       }
    }

    schedule();
    goto repeat;

#else
    FlushLRU(lru_handle);
    goto repeat;
#endif

#if (LINUX_20 | LINUX_22 | LINUX_24)

WaitOnBuffer:;
    lru->Waiters++;
    NWUnlockLRU(lru_handle);

    NWLockBuffer(lru);
    if (!(lru->state & L_DATAVALID) ||
	 (lru->block != block)      ||
	 (lru->volnum != volume->VolumeNumber))
       return 0;

    goto repeat;

#endif

#if (DOS_UTIL | LINUX_UTIL | WINDOWS_NT_UTIL)

WaitOnBuffer:;
    FlushLRU(lru_handle);

    if (retries++ > 5)
       return 0;

    goto repeat;
#endif

}

