
/***************************************************************************
*
*   Copyright (c) 1998, 1999 Timpanogas Research Group, Inc.
*   895 West Center Street
*   Orem, Utah  84057
*   jmerkey@timpanogas.com
*
*   This program is free software; you can redistribute it and/or modify it
*   under the terms of the GNU General Public License as published by the
*   Free Software Foundation, version 2, or any later version.
*
*   This program is distributed in the hope that it will be useful, but
*   WITHOUT ANY WARRANTY; without even the implied warranty of
*   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
*   General Public License for more details.
*
*   You are free to modify and re-distribute this program in accordance
*   with the terms specified in the GNU Public License.  The copyright
*   contained in this code is required to be present in any derivative
*   works and you are required to provide the source code for this
*   program as part of any commercial or non-commercial distribution.
*   You are required to respect the rights of the Copyright holders
*   named within this code.
*
*   jmerkey@timpanogas.com and TRG, Inc. are the official maintainers of
*   this code.  You are encouraged to report any bugs, problems, fixes,
*   suggestions, and comments about this software to jmerkey@timpanogas.com
*   or linux-kernel@vger.rutgers.edu.  New releases, patches, bug fixes, and
*   technical documentation can be found at www.timpanogas.com.  TRG will
*   periodically post new releases of this software to www.timpanogas.com
*   that contain bug fixes and enhanced capabilities.
*
*   Original Authorship      :
*      source code written by Jeff V. Merkey, TRG, Inc.
*
*   Original Contributors    :
*      Jeff V. Merkey, TRG, Inc.
*      Darren Major, TRG, Inc.
*      Alan Cox, RedHat Software, Inc.
*
****************************************************************************
*
*
*   AUTHOR   :  Jeff V. Merkey (jmerkey@timpanogas.com)
*   FILE     :  MMAP.C
*   DESCRIP  :  FENRIS VFS Memory Mapping Module for Linux
*   DATE     :  December 6, 1998
*
*
***************************************************************************/

#include "globals.h"

#if (LINUX_24)

// this function returns the volume logical block number from the
// logical file block number.

ULONG NWFileMapBlock(VOLUME *volume, HASH *hash, ULONG block, ULONG bsize)
{
    register ULONG Length, vBlock, BlocksPerCluster;
    register long FATChain, index, voffset;
    register ULONG StartIndex, StartOffset, rc;
    register FAT_ENTRY *FAT;
    SUBALLOC_MAP map;

    // if a subdirectory or bad bsize then return error
    if (!hash || (hash->Flags & SUBDIRECTORY_FILE) || !bsize)
       return (ULONG) -1;

    // get the proper namespace stream cluster and size
    switch (hash->NameSpace)
    {
       case DOS_NAME_SPACE:
       case MAC_NAME_SPACE:
	  FATChain = hash->FirstBlock;
	  Length = hash->FileSize;
	  break;

       default:
	  return (ULONG) -1;
    }

    StartIndex = (block * bsize) / volume->ClusterSize;
    StartOffset = (block * bsize) % volume->ClusterSize;
    BlocksPerCluster = volume->ClusterSize / bsize;
    
    // we always start with an index of zero
    index = 0;
    if (FATChain & 0x80000000)
    {
       // check for EOF
       if (FATChain == (ULONG) -1)
          return (ULONG) -1;

       // index must be null here
       if (StartIndex)
          return (ULONG) -1;

       rc = MapSuballocNode(volume, &map, FATChain);
       if (rc)
          return (ULONG) -1;

       if (StartOffset >= map.Size)
          return (ULONG) -1;

       if (StartOffset >= map.clusterSize[0])
       {
	  if (map.Count == 1)
             return (ULONG) -1;

          voffset = StartOffset - map.clusterSize[0];
	  vBlock = (map.clusterNumber[1] * BlocksPerCluster) +
	          ((map.clusterOffset[1] + voffset) / bsize);
	  return vBlock;
       }
       else
       {
	  vBlock = (map.clusterNumber[0] * BlocksPerCluster) +
		  ((map.clusterOffset[0] + StartOffset) / bsize);
	  return vBlock;
       }
    }

    FAT = GetFatEntry(volume, FATChain);
    if (FAT)
       index = FAT->FATIndex;

    while (FAT && FAT->FATCluster)
    {
       // we detected a hole in the file
       if (StartIndex < index)
       {
          // return 0 if we detect a hole in the file.
	  vBlock = 0;
	  return vBlock;
       }

       // we found our cluster in the chain
       if (StartIndex == index)
       {
	  vBlock = (FATChain * BlocksPerCluster) + (StartOffset / bsize);
	  return vBlock;
       }

       // bump to the next cluster
       FATChain = FAT->FATCluster;

       // check if the next cluster is a suballoc element or EOF marker
       if (FATChain & 0x80000000)
       {
	  // end of file
	  if (FATChain == (ULONG) -1)
             return (ULONG) -1;

	  // check for valid index
	  if ((index + 1) == StartIndex)
	  {
	     rc = MapSuballocNode(volume, &map, FATChain);
	     if (rc)
                return (ULONG) -1;

	     if (StartOffset >= map.Size)
                return (ULONG) -1;

             if (StartOffset >= map.clusterSize[0])
             {
	        if (map.Count == 1)
                   return (ULONG) -1;

                voffset = StartOffset - map.clusterSize[0];
	        vBlock = (map.clusterNumber[1] * BlocksPerCluster) +
	                ((map.clusterOffset[1] + voffset) / bsize);
	        return vBlock;
             }
             else
             {
	        vBlock = (map.clusterNumber[0] * BlocksPerCluster) +
		        ((map.clusterOffset[0] + StartOffset) / bsize);
	        return vBlock;
             }
	  }
          return (ULONG) -1;
       }

       // get next fat table entry and index
       FAT = GetFatEntry(volume, FATChain);
       if (FAT)
	  index = FAT->FATIndex;
    }
    return (ULONG) -1;

}

ULONG nwfs_map_extents(VOLUME *volume, HASH *hash, ULONG block, ULONG bsize,
		       ULONG *device, ULONG extend, ULONG *retCode,
		       ASYNCH_IO *io, struct inode *inode)
{
    register ULONG ccode, disk, vBlock, BlocksPerPage;
    register ULONG vOffset, vIndex;
    ULONG count = 0;
    DOS dos;
    nwvp_asynch_map map[8];

    if (!bsize)
       return 0;

    if (device)
       *device = 0;
    
    if (retCode)
       *retCode = 0;

    BlocksPerPage = (IO_BLOCK_SIZE / bsize);
    if (!BlocksPerPage)
       return 0;

    if (extend)
    {
       ccode = ReadDirectoryRecord(volume, &dos, hash->DirNo);
       if (ccode)
          return 0;
     
       // NWWriteFile() supports an extend mode that will extend the 
       // meta-data for a file if we pass a NULL address for the data 
       // buffer argument without writing any data to the extended 
       // area.

       ccode = NWWriteFile(volume,
		           &dos.FirstBlock, 
		           dos.Flags, 
		           block * bsize,
		           NULL,
		           bsize,
		           0,
		           0,
		           retCode,
		           KERNEL_ADDRESS_SPACE,
                           ((volume->VolumeFlags & SUB_ALLOCATION_ON) ? 1 : 0),
		           dos.FileAttributes);

       if (ccode != bsize)       
       {
#if (VERBOSE)
          NWFSPrint("nwfs:  out of space during extend\n");
#endif
	  return -ENOSPC;
       }

       if (retCode && *retCode)
       {
#if (VERBOSE)
          NWFSPrint("nwfs:  error code %d returned during extend\n", 
		   (int)*retCode);
#endif
	  return 0;
       }
       if (((block + 1) * bsize) > dos.FileSize)
          dos.FileSize = (block + 1) * bsize;

       if (inode)
          inode->i_size = dos.FileSize;

#if (HASH_FAT_CHAINS)
       hash->FileSize = dos.FileSize; 
       hash->FirstBlock = dos.FirstBlock;
#endif

       ccode = WriteDirectoryRecord(volume, &dos, hash->DirNo);
       if (ccode)
          return 0;
    }	    

    vBlock = NWFileMapBlock(volume, hash, block, bsize);
    if (!vBlock || (vBlock == (ULONG) -1))
    {
#if (VERBOSE)
       NWFSPrint("[%s] no map vBlock-%d block-%d\n", hash->Name, 
		 (int)vBlock, (int)block);
#endif
       return 0;
    }

    vIndex = vBlock / BlocksPerPage;
    vOffset = vBlock % BlocksPerPage;
    ccode = nwvp_vpartition_map_asynch_write(volume->nwvp_handle,
                                             vIndex, &count, &map[0]); 
    if (!ccode && count)
    {
       disk = map[0].disk_id; 
       if (SystemDisk[disk])
       { 
          if (device)
	  {
  	     *device = (ULONG)SystemDisk[disk]->PhysicalDiskHandle;
#if (VERBOSE)
             NWFSPrint("[%s]-%d block-%d vblock-%d lba-%d disk-%d dev-%04X\n", 
		    hash->Name,
		    (int)hash->FileSize,
		    (int)block,
		    (int)vBlock,
		    (int)(map[0].sector_offset + vOffset),
		    (int)map[0].disk_id,
		    (unsigned)*device);	     
#endif
	     return (ULONG)(map[0].sector_offset + vOffset);
	  }
       }
    }
    return 0;
}

int nwfs_get_block(struct inode *inode, long block,
                   struct buffer_head *bh_result, int create)
{
    register VOLUME *volume = inode->i_sb->u.generic_sbp;
    register HASH *hash = inode->u.generic_ip;
    register ULONG lba, BlocksPerCluster;
    ULONG device = 0, retCode = 0;
    
    if (!hash)
       return -EBADF;

    if (!volume)
       return -EBADF;

    if (!inode->i_sb->s_blocksize)
       return -EBADF;
    
    BlocksPerCluster = (volume->ClusterSize / inode->i_sb->s_blocksize);
    if ((block < 0) || (block > (volume->VolumeClusters * BlocksPerCluster)))
       return -EIO;

    NWLockFileExclusive(hash);

    lba = nwfs_map_extents(volume, hash, block, inode->i_sb->s_blocksize, 
		           &device, 0, &retCode, 0, inode);
    if (lba && device)
    {
       bh_result->b_dev = device;
       bh_result->b_blocknr = lba;
       bh_result->b_state |= (1UL << BH_Mapped);

       NWUnlockFile(hash);
       return 0;
    }

    if (!create)
    {
       NWUnlockFile(hash);
       return 0;
    }
	    
    lba = nwfs_map_extents(volume, hash, block, inode->i_sb->s_blocksize, 
		           &device, TRUE, &retCode, 0, inode);
    if (lba && device)
    {
       bh_result->b_dev = device;
       bh_result->b_blocknr = lba;
       bh_result->b_state |= (1UL << BH_Mapped);
       bh_result->b_state |= (1UL << BH_New);

       NWUnlockFile(hash);
       return 0;
    }

    NWUnlockFile(hash);
    return -EIO;
}

int nwfs_writepage(struct dentry *dentry, struct page *page)
{
    return block_write_full_page(page, nwfs_get_block);
}

int nwfs_readpage(struct dentry *dentry, struct page *page)
{
    return block_read_full_page(page, nwfs_get_block);
}

int nwfs_prepare_write(struct file *file, struct page *page, unsigned from, unsigned to)
{
    return block_prepare_write(page, from, to, nwfs_get_block);
}

int nwfs_bmap(struct address_space *mapping, long block)
{
    return generic_block_bmap(mapping, block, nwfs_get_block);
}

struct address_space_operations nwfs_aops =
{
    readpage:         nwfs_readpage,
    writepage:        nwfs_writepage,
    prepare_write:    nwfs_prepare_write,
    commit_write:     generic_commit_write,
    bmap:	      nwfs_bmap,
};

#endif

#if (LINUX_22)
#endif

#if (LINUX_20)
#endif
